/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

package fr.x9c.barista.api;

import java.io.InputStream;

import fr.x9c.cadmium.kernel.Block;
import fr.x9c.cadmium.kernel.Channel;
import fr.x9c.cadmium.kernel.Custom;
import fr.x9c.cadmium.kernel.Value;

/**
 * This class represents a Java class definition.
 *
 * @author <a href="mailto:barista@x9c.fr">Xavier Clerc</a>
 * @version 1.3
 * @since 1.0
 */
public final class ByteCode {

    /**
     * No instance of this class.
     */
    private ByteCode() {
    } // end empty constructor

    /**
     * Generates bytecode for a class definition.
     * @param v version usef for encoding - should not be <tt>null</tt>
     * @param cd class definition to encode - should not be <tt>null</tt>
     * @return the bytecode for the passed definition
     * @throws BaristaException if an error occurs
     */
    public static byte[] encode(final Version v,
                                final ClassDefinition cd)
        throws BaristaException {
        assert v != null : "null v";
        assert cd != null : "null cd";
        final Value res = 
            API.callback("fr.x9c.barista.api.ClassDefinition.encode",
                         v.toValue(),
                         cd.toValue());
        return res.asBlock().getBytes();
    } // end method 'encode(Version, ClassDefinition)'

    /**
     * Generates bytecode for a class definition.
     * @param cd class definition to encode - should not be <tt>null</tt>
     * @return the bytecode for the passed definition
     * @throws BaristaException if an error occurs
     */
    public static byte[] encode(final ClassDefinition cd)
        throws BaristaException {
        assert cd != null : "null cd";
        return encode(Version.Java_1_6, cd);
    } // end method 'encode(ClassDefinition)'

    /**
     * Constructs a class definition from an input stream.
     * @param in source for class bytecode - should not be <tt>null</tt>
     * @return the class definition of the passed bytecode
     * @throws BaristaException if an error occurs
     */
    public static ClassDefinition decode(final InputStream in)
        throws BaristaException {
        assert in != null : "null in";
        final Block b = Block.createCustom(Custom.CHANNEL_SIZE,
                                           Custom.CHANNEL_OPS);
        b.setCustom(new Channel(in));
        final Value res =
            API.callback("fr.x9c.barista.api.ClassDefinition.decode",
                         Value.createFromBlock(b));
        return ClassDefinition.fromValue(res);
    } // end method 'decode(InputStream)'

} // end class 'ByteCode'
