(*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *)

open BaristaLibrary.Utils

let fail () =
  print_string "KO";
  exit 1

let fail_if x y =
  if x <> y then fail ()

let char_tests () =
  print_string "char tests ... ";
  for i = 0 to 255 do
    let ch = Char.chr i in
    fail_if (char_of_uchar (uchar_of_char ch)) ch
  done;
  print_endline "OK"

let modified_utf8_tests () =
  print_string "modified utf8 tests ... ";
  let s = String.create 256 in
  for i = 0 to 255 do
    String.set s i (Char.chr i)
  done;
  let s2 = string_of_utf8 (utf8_of_string (s)) in
  fail_if s s2;
  let b = bytes_of_modified (modified_of_utf8 (utf8_of_string s)) in
  let s' = string_of_utf8 (utf8_of_modified (modified_of_bytes b)) in
  fail_if s s';
  print_endline "OK"

let utf8_tests () =
  let sgn x = if x = 0 then 0 else x / (abs x) in
  print_string "utf8 tests ... ";
  let s = utf8_of_string "azerty" in
  fail_if (utf8_length s) 6;
  if not (uchar_eq (utf8_get s 2) (uchar_of_char 'e')) then fail ();
  if not (utf8_eq s (utf8_of_string "azerty")) then fail ();
  if utf8_eq s (utf8_of_string "azertt") then fail ();
  fail_if
    (sgn (compare "aqw" "bfe"))
    (sgn (utf8_compare (utf8_of_string "aqw") (utf8_of_string "bfe")));
  fail_if
    (utf8_index_from (utf8_of_string "azerta") 1 (uchar_of_char 'a'))
    5;
  (try
    ignore (utf8_index_from (utf8_of_string "azerta") 1 (uchar_of_char 'h'));
    fail ()
  with Not_found -> ());
  fail_if
    (utf8_rindex_from (utf8_of_string "azerta") 4 (uchar_of_char 't'))
    4;
  (try
    ignore (utf8_rindex_from (utf8_of_string "azerta") 1 (uchar_of_char 't'));
    fail ()
  with Not_found -> ());
  if not (utf8_eq (utf8_of_string "zert") (utf8_substring s 1 4)) then fail ();
  fail_if
    (utf8_length (utf8_substring s 5 4))
    0;
  if not (utf8_eq s ((utf8_of_string "aze") ++ (utf8_of_string "rty"))) then fail ();
  if not (utf8_eq s (utf8_concat [(utf8_of_string "aze"); (utf8_of_string "rty")])) then fail ();
  print_endline "OK"

let () =
  char_tests ();
  modified_utf8_tests ();
  utf8_tests ()
