/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.FileOutputStream;
import java.util.Collections;
import java.util.List;
import static java.util.Arrays.asList;

import fr.x9c.barista.api.*;
import static fr.x9c.barista.api.Instruction.*;
import static tests.Common.*;

public final class Source {

    private static final FieldRef systOut =
	new FieldRef("java.lang.System",
		     "out",
		     new Descriptor.Class("java.io.PrintStream"));

    public static void main(final String[] args) throws Throwable {
	final List<Instruction> instructions =
	    asList(/* ddiv */
		   new GETSTATIC(systOut),
		   new LDC2_W_DOUBLE(7.5),
		   new LDC2_W_DOUBLE(2.5),
		   DDIV,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.DOUBLE))),

		   /* dmul */
		   new GETSTATIC(systOut),
		   new LDC2_W_DOUBLE(1.5),
		   new LDC2_W_DOUBLE(5.0),
		   DMUL,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.DOUBLE))),

		   /* dneg */
		   new GETSTATIC(systOut),
		   new LDC2_W_DOUBLE(1.5),
		   DNEG,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.DOUBLE))),

		   /* drem */
		   new GETSTATIC(systOut),
		   new LDC2_W_DOUBLE(5.0),
		   new LDC2_W_DOUBLE(1.5),
		   DREM,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.DOUBLE))),

		   /* dsub */
		   new GETSTATIC(systOut),
		   new LDC2_W_DOUBLE(5.0),
		   new LDC2_W_DOUBLE(1.5),
		   DSUB,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.DOUBLE))),

		   /* fdiv */
		   new GETSTATIC(systOut),
		   new LDC_FLOAT(7.5f),
		   new LDC_FLOAT(2.5f),
		   FDIV,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.FLOAT))),

		   /* fmul */
		   new GETSTATIC(systOut),
		   new LDC_FLOAT(1.5f),
		   new LDC_FLOAT(5.0f),
		   FMUL,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.FLOAT))),

		   /* fneg */
		   new GETSTATIC(systOut),
		   new LDC_FLOAT(1.5f),
		   FNEG,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.FLOAT))),

		   /* frem */
		   new GETSTATIC(systOut),
		   new LDC_FLOAT(5.0f),
		   new LDC_FLOAT(1.5f),
		   FREM,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.FLOAT))),

		   /* fsub */
		   new GETSTATIC(systOut),
		   new LDC_FLOAT(5.0f),
		   new LDC_FLOAT(1.5f),
		   FSUB,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.FLOAT))),

		   /* iand */
		   new GETSTATIC(systOut),
		   new LDC_INT(0xFC),
		   new LDC_INT(0xAB),
		   IAND,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* idiv */
		   new GETSTATIC(systOut),
		   new LDC_INT(24),
		   new LDC_INT(7),
		   IDIV,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* imul */
		   new GETSTATIC(systOut),
		   new LDC_INT(6),
		   new LDC_INT(7),
		   IMUL,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* ineg */
		   new GETSTATIC(systOut),
		   new LDC_INT(7),
		   INEG,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* ior */
		   new GETSTATIC(systOut),
		   new LDC_INT(4),
		   new LDC_INT(3),
		   IOR,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* irem */
		   new GETSTATIC(systOut),
		   new LDC_INT(24),
		   new LDC_INT(7),
		   IREM,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* ishl */
		   new GETSTATIC(systOut),
		   new LDC_INT(24),
		   new LDC_INT(2),
		   ISHL,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* ishr */
		   new GETSTATIC(systOut),
		   new LDC_INT(24),
		   new LDC_INT(2),
		   ISHR,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* isub */
		   new GETSTATIC(systOut),
		   new LDC_INT(24),
		   new LDC_INT(17),
		   ISUB,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* iushr */
		   new GETSTATIC(systOut),
		   new LDC_INT(24),
		   new LDC_INT(2),
		   IUSHR,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* ixor */
		   new GETSTATIC(systOut),
		   new LDC_INT(27),
		   new LDC_INT(8),
		   IXOR,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.INT))),

		   /* land */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(0xFCL),
		   new LDC2_W_LONG(0xABL),
		   LAND,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),

		   /* ldiv */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(24L),
		   new LDC2_W_LONG(7L),
		   LDIV,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),

		   /* lmul */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(6L),
		   new LDC2_W_LONG(7L),
		   LMUL,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),

		   /* lneg */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(7L),
		   LNEG,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),

		   /* lor */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(4L),
		   new LDC2_W_LONG(3L),
		   LOR,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),

		   /* lrem */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(24L),
		   new LDC2_W_LONG(7L),
		   LREM,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),

		   /* lshl */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(24L),
		   new LDC_INT(2),
		   LSHL,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),

		   /* lshr */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(24L),
		   new LDC_INT(2),
		   LSHR,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),

		   /* lsub */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(24L),
		   new LDC2_W_LONG(17L),
		   LSUB,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),

		   /* lushr */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(24L),
		   new LDC_INT(2),
		   LUSHR,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),

		   /* lxor */
		   new GETSTATIC(systOut),
		   new LDC2_W_LONG(27L),
		   new LDC2_W_LONG(8L),
		   LXOR,
		   new INVOKEVIRTUAL(new MethodRef("java.io.PrintStream",
						   "println",
						   Descriptor.VOID,
						   descs(Descriptor.LONG))),
		   RETURN);
	final Method main = compileMethod(asList(AccessFlag.Public, AccessFlag.Static),
					  "main",
					  Descriptor.VOID,
					  descs(new Descriptor.Array(new Descriptor.Class("java.lang.String"))),
					  5,
					  4,
					  Collections.<Attribute.ExceptionTableElement>emptyList(),
					  Collections.<Attribute>emptyList(),
					  Collections.<Attribute>emptyList(),
					  instructions);
	final ClassDefinition cd = compileClass(null,
						null,
						null,
						null,
						null,
						null,
						asList(main));
	final byte[] bc = ByteCode.encode(cd);
	FileOutputStream out = new FileOutputStream("pack/Test.class");
	out.write(bc);
	out.flush();
    }

}
