/*
 * This file is part of Barista.
 * Copyright (C) 2007-2010 Xavier Clerc.
 *
 * Barista is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * Barista is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

import java.io.FileOutputStream;
import java.util.Collections;
import java.util.List;
import static java.util.Arrays.asList;

import fr.x9c.barista.api.*;
import static fr.x9c.barista.api.Instruction.*;
import static tests.Common.*;

public final class Source {

    private static final FieldRef systOut =
	new FieldRef("java.lang.System",
		     "out",
		     new Descriptor.Class("java.io.PrintStream"));

    private static final MethodRef println =
	new MethodRef("java.io.PrintStream",
		      "println",
		      Descriptor.VOID,
		      descs(new Descriptor.Class("java.lang.String")));

    public static void main(final String[] args) throws Throwable {
	final Field x = new Field(asList(AccessFlag.Public),
				  "x",
				  new Descriptor.Class("java.lang.Number"),
				  asList(new Attribute.Signature(Attribute.Signature.Kind.FIELD,
                                                                 "TA;")));
	final List<Instruction> instructions =
	    asList(new GETSTATIC(systOut),
		   new LDC_STRING("hello\n..."),
		   new INVOKEVIRTUAL(println),
		   RETURN);
	final Annotation depr =
	    new Annotation("java.lang.Deprecated", 
			   Collections.<String>emptyList(),
			   Collections.<Annotation.ElementValue>emptyList());
	final Annotation annot =
	    new Annotation("pack.MyAnnotation",
			   asList("e", "c", "b", "a"),
			   asList(new Annotation.EnumValue("pack.MyAnnotation$E", "E3"),
				  new Annotation.ArrayValue(new Annotation.ElementValue[] {
					  new Annotation.IntValue(5),
					  new Annotation.IntValue(7)
				      }),
				  new Annotation.FloatValue(3.14f),
				  new Annotation.StringValue("xyz")));
	final List<Attribute> methAttrs =
	    asList((Attribute) new Attribute.RuntimeVisibleAnnotations(asList(annot, depr)));
	final Method main = compileMethod(asList(AccessFlag.Public, AccessFlag.Static),
					  "main",
					  Descriptor.VOID,
					  descs(new Descriptor.Array(new Descriptor.Class("java.lang.String"))),
					  4,
					  4,
					  Collections.<Attribute.ExceptionTableElement>emptyList(),
					  Collections.<Attribute>emptyList(),
					  methAttrs,
					  instructions);
	final ClassDefinition cd = compileClass(null,
						null,
						null,
						null,
						asList(x),
						asList((Attribute) new Attribute.SourceFile("<<Filename>>"),
						       (Attribute) new Attribute.Signature(Attribute.Signature.Kind.CLASS,
                                                                                           "<A:Ljava/lang/Number;>Ljava/lang/Object;")),
						asList(main));
	final byte[] bc = ByteCode.encode(cd);
	FileOutputStream out = new FileOutputStream("pack/Test.class");
	out.write(bc);
	out.flush();
    }

}
